#' Data
#'
#' Data encapsulation as a list (days) of lists (components).
#'
#' The private field .data is a list where each cell contains the variables for a period
#' of time of 24 hours, from time P+1 until P the next day where P is an integer between
#' 0 and 23. .data[[1]] refers to the first measured data (serie and exogenous
#' variables): the corresponding times can be accessed through the function
#' \code{getTime()} below. Each cell .data[[i]] is itself a list containing five slots,
#' as described in the 'field' section.
#'
#' @usage data <- Data$new()
#'
#' @field .data[[i]] List of
#' \itemize{
#'   \item time: vector of times
#'   \item centered_serie: centered serie
#'   \item level: corresponding level
#'   \item exo: exogenous variables
#'   \item exo_hat: predicted exogenous variables (for next day)
#' }
#'
#' @section Methods:
#' \describe{
#' \item{\code{getSize()}}{
#'   Number of series in dataset.}
#' \item{\code{getStdHorizon()}}{
#'   Number of time steps from serie[1] until midnight}
#' \item{\code{append(time, serie, exo, exo_hat)}}{
#'   Measured data for given vector of times + exogenous predictions from
#'   last midgnight.}
#' \item{\code{getTime(index)}}{
#'   Times (vector) at specified index.}
#' \item{\code{getCenteredSerie(index)}}{
#'   Centered serie at specified index.}
#' \item{\code{getCenteredSeries(indices)}}{
#'   Centered series at specified indices (in columns).}
#' \item{\code{getLevel(index)}}{
#'   Level at specified index.}
#' \item{\code{getSerie(index)}}{
#'   Serie (centered+level) at specified index.}
#' \item{\code{getSeries(indices)}}{
#'   Series at specified indices (in columns).}
#' \item{\code{getExoHat(index)}}{
#'   Predicted exogenous variables at specified index.}
#' \item{\code{getExo(index)}}{
#'   Measured exogenous variables at specified index.}
#' \item{\code{removeFirst()}}{
#'   Remove first list element (if truncated).}
#' \item{\code{removeLast()}}{
#'   Remove last list element (if truncated).}
#' }
#'
#' @docType class
#' @format R6 class
#'
Data = R6::R6Class("Data",
	private = list(
		.data = list()
	),
	public = list(
		getSize = function()
			length(private$.data)
		,
		getStdHorizon = function()
			24 - as.POSIXlt( private$.data[[1]]$time[1] )$hour + 1
		,
		append = function(time, serie, exo, exo_hat)
		{
			level = mean(serie, na.rm=TRUE)
			centered_serie = serie - level
			private$.data[[length(private$.data)+1]] <- list(
				"time"=time, #H-24 --> H-1
				"centered_serie"=centered_serie, #at 'time'
				"level"=level, #at 'time'
				"exo"=exo, #at 'time' (yersteday 0am to last midnight)
				"exo_hat"=exo_hat) #today 0am to next midnight
		},
		getTime = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$time
		},
		getCenteredSerie = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$centered_serie
		},
		getCenteredSeries = function(indices)
			sapply(indices, function(i) self$getCenteredSerie(i))
		,
		getLevel = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$level
		},
		getSerie = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$centered_serie + private$.data[[index]]$level
		},
		getSeries = function(indices)
			sapply(indices, function(i) self$getSerie(i))
		,
		getExoHat = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$exo_hat
		},
		getExo = function(index)
		{
			index = dateIndexToInteger(index, self)
			private$.data[[index]]$exo
		},
		removeFirst = function()
			private$.data <- private$.data[2:length(private$.data)]
		,
		removeLast = function()
			private$.data <- private$.data[1:(length(private$.data)-1)]
	)
)
